from typing import Dict, List

from rich.console import Console
from rich.table import Table


def render_config_table(configurations: List[Dict[str, str]]) -> None:
    """Renders a table using the supplied configuration."""
    table = Table()
    table.add_column("Name", style="white", no_wrap=True)
    table.add_column("Value", style="cyan")

    for configuration in configurations:
        table.add_row(configuration["name"], configuration["value"])

    console = Console()
    console.print(table)



import os

from decouple import AutoConfig, Csv

# configures decouple to use settings.ini or .env file from another directory
if REVIEWS_PATH_TO_CONFIG := os.environ.get("REVIEWS_PATH_TO_CONFIG", None):
    config = AutoConfig(search_path=REVIEWS_PATH_TO_CONFIG)
else:
    config = AutoConfig()

# Github Config
GITHUB_TOKEN = config("GITHUB_TOKEN", cast=str, default="")
GITHUB_USER = config("GITHUB_USER", cast=str, default="")
GITHUB_URL = config("GITHUB_URL", cast=str, default="https://api.github.com")
GITHUB_DEFAULT_PAGE_SIZE = config("GITHUB_DEFAULT_PAGE_SIZE", cast=int, default=100)

# Gitlab config
GITLAB_TOKEN = config("GITLAB_TOKEN", cast=str, default="")
GITLAB_USER = config("GITLAB_USER", cast=str, default="")
GITLAB_URL = config("GITLAB_URL", cast=str, default="https://gitlab.com")
GITLAB_DEFAULT_PAGE_SIZE = config("GITLAB_DEFAULT_PAGE_SIZE", cast=int, default=100)

# Application Config
REVIEWS_PATH_TO_CONFIG = config("REVIEWS_PATH_TO_CONFIG", cast=str, default=None)
REVIEWS_DELAY_REFRESH = config("REVIEWS_DELAY_REFRESH", cast=int, default=60)
REVIEWS_GITHUB_REPOSITORY_CONFIGURATION = config(
    "REVIEWS_GITHUB_REPOSITORY_CONFIGURATION",
    cast=Csv(),
    default="apoclyps/reviews",
)
REVIEWS_GITLAB_REPOSITORY_CONFIGURATION = config(
    "REVIEWS_GITLAB_REPOSITORY_CONFIGURATION",
    cast=Csv(),
    default="27629846:apoclyps/reviews",
)
REVIEWS_LABEL_CONFIGURATION = config(
    "REVIEWS_LABEL_CONFIGURATION",
    cast=Csv(),
    default="blocked/orange,docker/blue,security/red,python/green",
)



from datetime import datetime
from time import sleep
from typing import List, Tuple

from rich.console import RenderGroup
from rich.live import Live
from rich.panel import Panel

from . import config
from .config import render_config_table
from .controller import GithubPullRequestController, GitlabPullRequestController, PullRequestController
from .layout import (
    RenderLayoutManager,
    generate_layout,
    generate_log_table,
    generate_progress_tracker,
    generate_tree_layout,
)

logs: List[Tuple[str, str]] = []


def add_log_event(message: str) -> List[Tuple[str, str]]:
    """adds a log event to a list of logs and displays the top 20."""
    global logs

    logs = logs[-20:]
    logs.append((str(datetime.now().strftime("%Y-%m-%d %H:%M:%S")), f"[white]{message}"))
    return logs


def single_render(provider: str) -> None:
    """Renders the Terminal UI Dashboard once before closing the application"""

    configuration = None
    body = None

    if provider == "gitlab":
        configuration = config.get_configuration(config=config.REVIEWS_GITLAB_REPOSITORY_CONFIGURATION)
        body = GitlabPullRequestController().render(configuration=configuration)
    else:
        configuration = config.get_configuration(config=config.REVIEWS_GITHUB_REPOSITORY_CONFIGURATION)
        body = GithubPullRequestController().render(configuration=configuration)

    layout_manager = RenderLayoutManager(layout=generate_layout(log=False, footer=False))
    layout_manager.render_layout(
        progress_table=None,
        body=body,
        pull_request_component=generate_tree_layout(configuration=configuration),
        log_component=None,
    )

    with Live(
        renderable=layout_manager.layout,
        refresh_per_second=5,
        transient=False,
        screen=False,
    ):
        add_log_event(message="updated")


def render(provider: str) -> None:
    """Renders Terminal UI Dashboard"""
    (
        job_progress,
        overall_progress,
        overall_task,
        progress_table,
    ) = generate_progress_tracker()

    # initial load should be from database
    add_log_event(message="initializing...")

    configuration = None
    controller: PullRequestController

    if provider == "gitlab":
        configuration = config.get_configuration(config=config.REVIEWS_GITLAB_REPOSITORY_CONFIGURATION)
        controller = GitlabPullRequestController()
    else:
        configuration = config.get_configuration(config=config.REVIEWS_GITHUB_REPOSITORY_CONFIGURATION)
        controller = GitlabPullRequestController()
        controller = GithubPullRequestController()

    layout_manager = RenderLayoutManager(layout=generate_layout())
    layout_manager.render_layout(
        progress_table=progress_table,
        body=Panel(
            RenderGroup(),
            title="Activity",
            border_style="blue",
        ),
        pull_request_component=generate_tree_layout(configuration=configuration),
        log_component=generate_log_table(logs=logs),
    )

    with Live(layout_manager.layout, refresh_per_second=5, screen=True):
        while True:
            if not overall_progress or overall_progress.finished:
                (
                    job_progress,
                    overall_progress,
                    overall_task,
                    progress_table,
                ) = generate_progress_tracker()

            add_log_event(message="waiting...")

            # update view (blocking operation)
            layout_manager.render_layout(
                progress_table=progress_table,
                body=controller.render(configuration=configuration),
                pull_request_component=generate_tree_layout(configuration=configuration),
                log_component=generate_log_table(logs=logs),
            )

            delay = config.REVIEWS_DELAY_REFRESH * 0.01
            while not overall_progress.finished:
                sleep(delay)
                for job in job_progress.tasks:
                    if not job.finished:
                        job_progress.advance(job.id)

                completed = sum(task.completed for task in job_progress.tasks)
                overall_progress.update(overall_task, completed=completed)

            add_log_event(message="updated")


def render_config(show: bool) -> None:
    """Renders a table displaying configuration used by Reviews."""
    configurations = [
        {
            "name": "GITHUB_TOKEN",
            "value": config.GITHUB_TOKEN if show else "".join("*" for _ in range(len(config.GITHUB_TOKEN))),
        },
        {"name": "GITHUB_USER", "value": config.GITHUB_USER},
        {"name": "GITHUB_URL", "value": config.GITHUB_URL},
        {
            "name": "GITLAB_TOKEN",
            "value": config.GITLAB_TOKEN if show else "".join("*" for _ in range(len(config.GITLAB_TOKEN))),
        },
        {"name": "GITLAB_USER", "value": config.GITLAB_USER},
        {"name": "GITLAB_URL", "value": config.GITLAB_URL},
        {"name": "REVIEWS_PATH_TO_CONFIG", "value": f"{config.REVIEWS_PATH_TO_CONFIG}"},
        {
            "name": "GITHUB_DEFAULT_PAGE_SIZE",
            "value": f"{config.GITHUB_DEFAULT_PAGE_SIZE}",
        },
        {"name": "REVIEWS_DELAY_REFRESH", "value": f"{config.REVIEWS_DELAY_REFRESH}"},
        {
            "name": "REVIEWS_GITHUB_REPOSITORY_CONFIGURATION",
            "value": ", ".join(config.REVIEWS_GITHUB_REPOSITORY_CONFIGURATION),
        },
        {
            "name": "REVIEWS_GITLAB_REPOSITORY_CONFIGURATION",
            "value": ", ".join(config.REVIEWS_GITLAB_REPOSITORY_CONFIGURATION),
        },
        {
            "name": "REVIEWS_LABEL_CONFIGURATION",
            "value": ", ".join(config.REVIEWS_LABEL_CONFIGURATION),
        },
    ]

    render_config_table(configurations=configurations)



from typing import Dict, List, Tuple

from decouple import Csv
from rich.color import ANSI_COLOR_NAMES

from .settings import REVIEWS_LABEL_CONFIGURATION


def get_configuration(config: Csv) -> List[Tuple[str, str]]:
    """converts a comma separated list of organizations/repositories into a list
    of tuples.
    """

    def _to_tuple(values: List[str]) -> Tuple[str, str]:
        return (values[0], values[1])

    def split(configuration: str) -> List[str]:
        if ":" in configuration:
            return configuration.split(sep=":", maxsplit=1)

        return configuration.split(sep="/", maxsplit=1)

    return [_to_tuple(values=split(configuration=configuration)) for configuration in config]


def get_label_colour_map() -> Dict[str, str]:
    """converts a comma separated list of organizations/repositories into a list
    of tuples.
    """

    def _preproc(label_colour: str) -> List[str]:
        return label_colour.lower().split(sep="/")

    return {
        label: f"[{colour}]"
        for label, colour in map(_preproc, REVIEWS_LABEL_CONFIGURATION)
        if colour in ANSI_COLOR_NAMES
    }



from typing import List

from github import Github
from github.GithubException import BadCredentialsException, UnknownObjectException
from github.PullRequest import PullRequest
from github.Repository import Repository
from gitlab import Gitlab
from gitlab.v4.objects.merge_requests import MergeRequest as GitlabMergeRequest
from gitlab.v4.objects.projects import Project as GitlabRepository

from .. import config
from ..errors import InvalidGithubToken, InvalidGitlabToken, RepositoryDoesNotExist


class GithubAPI:
    """Create and execute requests using the Github API"""

    def __init__(self) -> None:
        self._client = Github(
            config.GITHUB_TOKEN,
            per_page=config.GITHUB_DEFAULT_PAGE_SIZE,
            base_url=config.GITHUB_URL,
        )

    def get_repository(self, org: str, repo: str) -> Repository:
        """Returns a repository for a given organization."""
        try:
            return self._client.get_repo(f"{org}/{repo}")
        except UnknownObjectException:
            raise RepositoryDoesNotExist(f"{org}/{repo} does not exist")
        except BadCredentialsException:
            raise InvalidGithubToken("GITHUB_TOKEN is not valid.")

    @staticmethod
    def _get_pull_requests(
        repository: Repository,
        state: str = "open",
        sort: str = "created",
    ) -> List[PullRequest]:
        return list(repository.get_pulls(state=state, sort=sort))

    def get_pull_requests(self, org: str, repo: str) -> List[PullRequest]:
        """Returns a list of pull requests for a given organization and repository."""
        repository = self.get_repository(org=org, repo=repo)
        return self._get_pull_requests(repository=repository)


class GitlabAPI:
    """Create and execute requests using the Gitlab API"""

    def __init__(self) -> None:
        self._client = Gitlab(url=config.GITLAB_URL, private_token=config.GITLAB_TOKEN)

    def get_repository(self, project_id: str, namespace: str) -> GitlabRepository:
        """Returns a repository for a given organization."""
        try:
            return self._client.projects.get(id=project_id)  # type: ignore
        except UnknownObjectException:
            raise RepositoryDoesNotExist(f"{project_id} does not exist for {namespace}")
        except BadCredentialsException:
            raise InvalidGitlabToken("GITLAB_TOKEN is not valid.")

    @staticmethod
    def _get_pull_requests(
        repository: Repository,
        state: str = "open",
        sort: str = "created",
    ) -> List[PullRequest]:
        return list(repository.get_pulls(state=state, sort=sort))

    def get_pull_requests(self, project_id: str, namespace: str) -> List[GitlabMergeRequest]:
        """Returns a list of pull requests for a given organization and repository."""
        repository = self.get_repository(project_id=project_id, namespace=namespace)
        return repository.mergerequests.list(state="opened", order_by="created_at", sort="asc")  # type: ignore



from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Dict, List, Optional

import humanize


@dataclass
class Label:
    """Labels on a Pull Request."""

    name: str


@dataclass
class PullRequest:
    """Models a Pull Request"""

    number: int
    title: str
    draft: bool
    repository_url: str
    link: str
    created_at: datetime
    updated_at: datetime
    approved: str
    approved_by_others: bool
    additions: int = 0
    deletions: int = 0
    labels: List[Label] = field(default_factory=list)

    def render_approved(self) -> str:
        """Renders the approved status as a colourised string"""
        status = ""

        if self.approved == "AUTHOR":
            status = "[grey]Author[/]"
        elif self.approved == "APPROVED":
            status = "[green]Approved[/]"
        elif self.approved == "CHANGES_REQUESTED":
            status = "[red]Changes Requested[/]"

        return status

    def render_approved_by_others(self) -> str:
        """Renders approved_by_others flag as a colourised string"""
        return "[green]Ready[/]" if self.approved_by_others else ""

    def render_labels(self, colour_map: Dict[str, str]) -> str:
        """Renders the labels as a joined colourised string"""
        return ", ".join([f"{colour_map.get(label.name.lower(), '[white]')}{label.name}[/]" for label in self.labels])

    def render_title(self) -> str:
        """Renders the title as a colourised string"""
        colour = ""
        title = self.title

        if title.startswith("[Security]"):
            title = title.removeprefix("[Security]")
            colour = "[bold red][Security][/]"

        if self.draft:
            colour += "[bold grey][Draft][/] "

        return f"{colour}[white][link={self.link}]{title}[/link][/]"

    def render_updated_at(self, since: Optional[datetime] = None) -> str:
        """Renders the updated_at as a human-readable and colourised string"""
        colour = ""
        suffix = ""

        now = since or datetime.now(timezone.utc)
        days = (now - self.updated_at).days

        if days >= 7:
            colour = "[red]"
        elif days >= 1:
            colour = "[yellow]"

        if colour:
            suffix = "[/]"

        return f"{colour}{humanize.naturaltime(self.updated_at, when=now)}{suffix}"

    def render_diff(self) -> str:
        """Renders the additions and deletions using the Github convention of +/-"""
        return f"[green]+{self.additions}[/green] [red]-{self.deletions}[/red]"



from typing import Optional

from rich.console import Console, RenderGroup
from rich.layout import Layout
from rich.panel import Panel
from rich.table import Table

from ..layout.components import Header

console = Console()


class RenderLayoutManager:
    """Renders the Layout for the Terminal UI"""

    def __init__(self, layout: Layout) -> None:
        self.layout = layout

    def render_layout(
        self,
        body: Panel,
        pull_request_component: RenderGroup,
        log_component: Optional[Table],
        progress_table: Optional[Table],
    ) -> Layout:
        """Renders the entire layout"""
        self.render_header()
        self.render_body(component=body)
        if log_component:
            self.render_log(component=log_component)
        self.render_configuration(component=pull_request_component)
        if progress_table:
            self.render_footer(progress_table=progress_table)

        return self.layout

    def render_header(self) -> None:
        """Renders the Header component"""
        self.layout["header"].update(Header())

    def render_body(self, component: Panel) -> None:
        """Renders the Body component"""
        self.layout["body"].update(component)

    def render_configuration(self, component: RenderGroup) -> None:
        """Renders the Main Body component"""
        self.layout["configuration"].update(Panel(component, title="Configuration", border_style="blue"))

    def render_log(self, component: Table) -> None:
        """Rends the log component"""
        self.layout["log"].update(Panel(component))

    def render_footer(self, progress_table: Table) -> None:
        """Renders the Footer component"""
        self.layout["footer"].update(progress_table)



from datetime import datetime, timezone
from typing import Dict, List, Tuple, Union

from github.PullRequest import PullRequest as ghPullRequest
from gitlab.v4.objects.merge_requests import MergeRequest as GitlabMergeRequest
from rich.console import RenderGroup
from rich.panel import Panel
from rich.table import Table

from reviews.source_control.client import GitlabAPI

from . import config
from .errors import RepositoryDoesNotExist
from .layout import render_pull_request_table, render_repository_does_not_exist
from .source_control import GithubAPI, Label, PullRequest


class PullRequestController:
    """Abstract base class for providier specific pull request controllers"""

    def render(self, configuration: List[Tuple[str, str]]) -> Panel:
        """Renders all pull requests for the provided configuration"""
        raise NotImplementedError("the render method needs to be implemented for your specified provider")


class GithubPullRequestController(PullRequestController):
    """retrieve and store pull requests."""

    def __init__(self) -> None:
        self.client = GithubAPI()

    def retrieve_pull_requests(self, org: str, repository: str) -> Union[Table, None]:
        """Renders Terminal UI Dashboard"""

        title = f"{org}/{repository}"
        pull_requests = []
        try:
            pull_requests = self.update_pull_requests(org=org, repository=repository)
        except RepositoryDoesNotExist:
            return render_repository_does_not_exist(
                title=title,
                link=f"https://www.github.com/{org}/{repository}",
            )

        if not pull_requests:
            return None

        return render_pull_request_table(
            title=title,
            pull_requests=pull_requests,
        )

    def update_pull_requests(self, org: str, repository: str) -> List[PullRequest]:
        """Updates repository models."""

        def _get_reviews(pull_request: ghPullRequest) -> Dict[str, str]:
            """Inner function to retrieve reviews for a pull request"""
            reviews = pull_request.get_reviews()
            res, seen = {}, []

            for review in reviews:
                if review.user.login in seen or review.state == "COMMENTED":
                    continue
                res[review.user.login] = review.state
                seen.append(review.user.login)
            return res

        pull_requests = self.client.get_pull_requests(org=org, repo=repository)

        code_review_requests = []
        for pull_request in pull_requests:

            reviews = _get_reviews(pull_request=pull_request)

            if pull_request.user.login == config.GITHUB_USER:
                approved_by_me = "AUTHOR"
            else:
                approved_by_me = reviews.get(config.GITHUB_USER, "")  # NOQA: R1721

            approved_by_others = any(
                [True for user, status in reviews.items() if user != config.GITHUB_USER and status == "APPROVED"]
            )
            labels = [Label(name=label.name) for label in pull_request.get_labels() if label.name]

            code_review_requests.append(
                PullRequest(
                    number=pull_request.number,
                    title=pull_request.title,
                    draft=pull_request.draft,
                    repository_url=f"https://www.github.com/{org}/{repository}",
                    link=f"https://www.github.com/{org}/{repository}/pull/{pull_request.number}",
                    additions=pull_request.additions,
                    deletions=pull_request.deletions,
                    created_at=pull_request.created_at.astimezone(tz=timezone.utc),
                    updated_at=pull_request.updated_at.astimezone(tz=timezone.utc),
                    approved=approved_by_me,
                    approved_by_others=approved_by_others,
                    labels=labels,
                )
            )

        return code_review_requests

    def render(self, configuration: List[Tuple[str, str]]) -> Panel:
        """Renders all pull requests for the provided configuration"""

        tables = [self.retrieve_pull_requests(org=org, repository=repo) for (org, repo) in configuration]

        # filter unrenderable `None` results
        return Panel(
            RenderGroup(*[t for t in tables if t]),
            title="Activity",
            border_style="blue",
        )


class GitlabPullRequestController(PullRequestController):
    """retrieve and store pull requests."""

    def __init__(self) -> None:
        self.client = GitlabAPI()

    def retrieve_pull_requests(self, project_id: str, namespace: str) -> Union[Table, None]:
        """Renders Terminal UI Dashboard"""
        pull_requests = []
        try:
            pull_requests = self.update_pull_requests(project_id=project_id, namespace=namespace)
        except RepositoryDoesNotExist:
            return render_repository_does_not_exist(
                title=namespace,
                link=f"https://gitlab.com/{namespace}",
            )

        if not pull_requests:
            return None

        return render_pull_request_table(
            title=namespace,
            pull_requests=pull_requests,
        )

    def update_pull_requests(self, project_id: str, namespace: str) -> List[PullRequest]:
        """Updates repository models."""

        def _get_reviews(pull_request: GitlabMergeRequest) -> Dict[str, str]:
            """Inner function to retrieve reviews for a pull request"""
            reviews = pull_request.approvals.get()

            return {reviewer["user"]["username"]: "approved" for reviewer in reviews.approvers}

        # ProjectMergeRequest
        pull_requests = self.client.get_pull_requests(project_id=project_id, namespace=namespace)

        code_review_requests = []
        for pull_request in pull_requests:

            reviews = _get_reviews(pull_request=pull_request)

            if pull_request.author["username"] == config.GITLAB_USER:
                approved_by_me = "AUTHOR"
            else:
                approved_by_me = reviews.get(config.GITLAB_USER, "")  # NOQA: R1721

            approved_by_others = any(
                [True for user, status in reviews.items() if user != config.GITLAB_USER and status == "APPROVED"]
            )

            def get_labels(labels: List[str]) -> List[Label]:
                visible_labels = []
                labels_mapping = {label: len(label) for label in labels}

                current_size = 0
                max_size = 18
                hidden_labels = 0
                for label, size in labels_mapping.items():
                    if current_size == 0:
                        current_size += size
                        visible_labels.append(label)
                        continue

                    if current_size + size < max_size:
                        current_size += size
                        visible_labels.append(label)
                    else:
                        hidden_labels += 1

                labels = [Label(name=label) for label in visible_labels]

                if hidden_labels:
                    labels.append(Label(name=f"+{hidden_labels} others"))

                return labels

            labels = get_labels(labels=pull_request.labels)

            link = pull_request.web_url

            code_review_requests.append(
                PullRequest(
                    number=pull_request.iid,
                    title=pull_request.title,
                    draft=pull_request.draft,
                    additions=0,
                    deletions=0,
                    link=link,
                    repository_url=link.split("/-/")[0],
                    created_at=datetime.strptime(pull_request.created_at, "%Y-%m-%dT%H:%M:%S.%f%z"),
                    updated_at=datetime.strptime(pull_request.updated_at, "%Y-%m-%dT%H:%M:%S.%f%z"),
                    approved=approved_by_me,
                    approved_by_others=approved_by_others,
                    labels=labels,
                )
            )

        return code_review_requests

    def render(self, configuration: List[Tuple[str, str]]) -> Panel:
        """Renders all pull requests for the provided configuration"""

        tables = [
            self.retrieve_pull_requests(project_id=project_id, namespace=namespace)
            for (project_id, namespace) in configuration
        ]

        # filter unrenderable `None` results
        return Panel(
            RenderGroup(*[t for t in tables if t]),
            title="Activity",
            border_style="blue",
        )
